from dataclasses import dataclass
from enum import Enum, auto
from typing import TypedDict, NewType

from ptyx_mcq.tools.config_parser import (
    DocumentId,
    StudentId,
    StudentName,
    OriginalQuestionAnswersDict,
    ApparentQuestionNumber,
    OriginalQuestionNumber,
    OriginalAnswerNumber,
)

Page = NewType("Page", int)


class DetectionStatus(Enum):
    CHECKED = auto()
    UNCHECKED = auto()
    PROBABLY_CHECKED = auto()
    PROBABLY_UNCHECKED = auto()

    def __repr__(self):
        return self.name


class RevisionStatus(Enum):
    MARKED_AS_CHECKED = auto()
    MARKED_AS_UNCHECKED = auto()

    def __repr__(self):
        return self.name


@dataclass(kw_only=True)
class PicData:
    # ID of the document:
    doc_id: DocumentId
    # page number:
    page: Page
    name: StudentName
    student_ID: StudentId
    # answers checked by the student for each question:
    answered: OriginalQuestionAnswersDict
    # Position of each checkbox in the page:
    positions: dict[tuple[OriginalQuestionNumber, OriginalAnswerNumber], tuple[int, int]]
    cell_size: int
    # Translation table ({question number before shuffling: after shuffling})
    questions_nums_conversion: dict[OriginalQuestionNumber, ApparentQuestionNumber]
    needs_review: bool
    detection_status: dict[tuple[OriginalQuestionNumber, OriginalAnswerNumber], DetectionStatus]
    revision_status: dict[tuple[OriginalQuestionNumber, OriginalAnswerNumber], RevisionStatus]
    pic_path: str


class DocumentData(TypedDict):
    pages: dict[Page, PicData]
    name: StudentName
    student_ID: StudentId
    answered: OriginalQuestionAnswersDict  # {question: set of answers}
    score: float
    score_per_question: dict[OriginalQuestionNumber, float]  # {question: score}



r"""
ptyx MCQ

This extension enables computer corrected quizzes.

An example:

    #LOAD{mcq}
    #SEED{8737545887}

    ===========================
    sty=my_custom_sty_file
    scores=1 0 0
    mode=all
    ids=~/my_students.csv
    ===========================


    <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

    ======= Mathematics ===========

    * 1+1 =
    - 1
    + 2
    - 3
    - 4

    - another answer

    ======= Litterature ==========

    * "to be or not to be", but who actually wrote that ?
    + W. Shakespeare
    - I. Newton
    - W. Churchill
    - Queen Victoria
    - Some bloody idiot

    * Jean de la Fontaine was a famous French
    - pop singer
    - dancer
    + writer
    - detective
    - cheese maker

    > his son is also famous for
    @{\color{blue}%s}
    - dancing french cancan
    - conquering Honolulu
    - walking for the first time on the moon
    - having breakfast at Tiffany

    + none of the above is correct

    >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>


One may include some PTYX code of course.

    """
import re

from ptyx.extensions import extended_python
from ptyx.latex_generator import Compiler
from ptyx_mcq.tools.io_tools import print_error

from .make.extend_latex_generator import MCQLatexGenerator
from .make.generate_ptyx_code import generate_ptyx_code
from .tools.include_parser import IncludeParser

# Note for closing tags:
# '@END' means closing tag #END must be consumed, unlike 'END'.
# So, use '@END_QUESTIONS_BLOCK' to close QUESTIONS_BLOCK,
# but use 'END_QUESTIONS_BLOCK' to close QUESTION, since
# #END_QUESTIONS_BLOCK must not be consumed then (it must close
# QUESTIONS_BLOCK too).


__tags__ = {
    # Tags used to structure MCQ
    "QCM": (0, 0, ["@END_QCM"]),
    "SECTION": (0, 0, ["SECTION", "END_QCM"]),
    "NEW_QUESTION": (0, 0, ["NEW_QUESTION", "CONSECUTIVE_QUESTION", "SECTION", "END_QCM"]),
    "CONSECUTIVE_QUESTION": (0, 0, ["NEW_QUESTION", "CONSECUTIVE_QUESTION", "SECTION", "END_QCM"]),
    "VERSION": (1, 0, ["VERSION", "NEW_QUESTION", "CONSECUTIVE_QUESTION", "SECTION", "END_QCM"]),
    "ANSWERS_BLOCK": (0, 0, ["@END_ANSWERS_BLOCK"]),
    "NEW_ANSWER": (2, 0, ["NEW_ANSWER", "END_ANSWERS_BLOCK"]),
    "ANSWERS_LIST": (2, 0, None),
    # Other tags
    "QCM_HEADER": (2, 0, None),
    "QUESTION_CONFIG": (1, 0, None),
    "DEBUG_MCQ": (0, 0, None),
    # Deprecated tags
    "L_ANSWERS": (1, 0, None),
}
__latex_generator_extension__ = MCQLatexGenerator


def autodetect_smallgraphlib(text: str) -> list[str]:
    smallgraphlib_detected = (
        re.search("(^import smallgraphlib)|(^from smallgraphlib import)", text, re.MULTILINE) is not None
    )
    if smallgraphlib_detected:
        try:
            # noinspection PyUnresolvedReferences
            from smallgraphlib.tikz_export import TikzPrinter

            preamble_additions = TikzPrinter.latex_preamble_additions()
            preamble_additions.remove(r"\usepackage{tikz}")
            return preamble_additions
        except ImportError:
            print_error(
                "This file tries to import `smallgraphlib` library, but it is not installed.\n"
                "You can install it with the following command:\npip install smallgraphlib"
            )
    return []


def main(text: str, compiler: Compiler) -> str:
    # Generation algorithm is the following:
    # 1. Parse AutoQCM code, to convert it to plain pTyX code.
    #    Doing this, we now know the number of questions, the number
    #    of answers per question and the students names.
    #    However, we can't know for know the number of the correct answer for
    #    each question, since questions numbers and answers numbers too will
    #    change during shuffling, when compiling pTyX code (and keeping track of
    #    them through shuffling is not so easy).
    # 2. Generate syntax tree, and then compile pTyX code many times to generate
    #    one test for each student. For each compilation, keep track of correct
    #    answers.
    #    All those data are stored in `latex_generator.mcq_data['answers']`.
    #    `latex_generator.mcq_data['answers']` is a dict
    #    with the following structure:
    #    {1:  [          <-- test n°1 (test id is stored in NUM)
    #         [0,3,5],   <-- 1st question: list of correct answers
    #         [2],       <-- 2nd question: list of correct answers
    #         [1,5],     ...
    #         ],
    #     2:  [          <-- test n°2
    #         [2,3,4],   <-- 1st question: list of correct answers
    #         [0],       <-- 2nd question: list of correct answers
    #         [1,2],     ...
    #         ],
    #    }

    try:
        text = IncludeParser(compiler.dir_path).parse(text, strict=True)
    except FileNotFoundError:
        print_error(
            "File not found when trying to resolve inclusions (see error message above).\n"
            f"Hint: command `mcq update-include {getattr(compiler.file_path, 'name', 'FILENAME')}` "
            "may fix it."
        )
    additional_header_lines = autodetect_smallgraphlib(text)

    # Call extended_python extension.
    text = extended_python.main(text, compiler)

    code = generate_ptyx_code(text, additional_header_lines=additional_header_lines)
    assert isinstance(code, str)
    return code



from dataclasses import dataclass
from math import nan, isnan

from ptyx_mcq.tools.config_parser import OriginalAnswerNumber


@dataclass
class AnswersData:
    checked: set[OriginalAnswerNumber]
    correct: set[OriginalAnswerNumber]
    all: set[OriginalAnswerNumber]

    @property
    def unchecked(self):
        return self.all - self.checked

    @property
    def incorrect(self):
        return self.all - self.correct

    @property
    def correctly_answered(self):
        return self.correct & self.checked | self.incorrect & self.unchecked

    @property
    def incorrectly_answered(self):
        return self.all - self.correctly_answered


@dataclass
class ScoreData:
    correct: float
    skipped: float
    incorrect: float = nan

    def __post_init__(self):
        if isnan(self.incorrect):
            self.incorrect = -self.correct


class EvaluationStrategies:
    """This class provides a collection of evaluation strategies.

     Those will be used to calculate the score for each question.

    Each strategy should be defined as a static method within the `EvaluationStrategies` class,
    with two required parameters:
    - the answers data (of type `AnswersData`),
    - the scores configuration (of type `ScoreData`).

    Note that only static methods within the `EvaluationStrategies` class will be recognized as
    evaluation strategies, so you may safely define auxiliary class methods without interfering.
    """

    @classmethod
    def get_all_strategies(cls) -> list[str]:
        return [name for name, func in vars(cls).items() if isinstance(func, staticmethod)]

    @staticmethod
    def some(answers: AnswersData, score: ScoreData) -> float:
        """If any of the correct answers have been selected and no incorrect answers
        have been selected, the function returns the maximal score.
        Otherwise, it returns the minimal score.

        The maximal score and minimal score are defined in the header of the ptyx file
        using the keywords "correct" and "incorrect", respectively."""
        # Answer is valid if and only if :
        # (proposed ≠ ∅ and proposed ⊆ correct) or (proposed = correct = ∅)
        ok = (answers.checked and answers.checked.issubset(answers.correct)) or (
            not answers.checked and not answers.correct
        )
        if ok:
            return score.correct
        elif not answers.checked:
            return score.skipped
        else:
            return score.incorrect

    @staticmethod
    def all(answers: AnswersData, score: ScoreData) -> float:
        """This function returns either the maximal score or the minimal score,
        depending on whether all correct answers are checked or not.

        If all correct answers are checked, the function returns the maximal score.
        Otherwise, it returns the minimal score."""
        ok = answers.checked == answers.correct
        if ok:
            return score.correct
        elif not answers.checked:
            return score.skipped
        else:
            return score.incorrect

    @staticmethod
    def proportional(answers: AnswersData, score: ScoreData) -> float:
        """The score returned by this function is calculated as a weighted average
        of the minimal and maximal scores.

        The minimal and maximal scores are specified in the header of the ptyx file
        using the keywords "correct" and "incorrect", respectively.

        The formula for the calculated score is `(1 - k) * minimal_score + k * maximal_score`,
        where k is the proportion of correct answers.
        This means that if all answers are correct, the score will be equal to the maximal score;
        if all answers are incorrect, the score will be equal to the minimal score;
        and if some answers are correct and others are incorrect, the score will be somewhere in between.

        To determine the value of k, the algorithm looks at the answers provided and checks
        which boxes were left blank when the proposed answer was incorrect,
        and which boxes were checked when the proposed answer was correct.
        The proportion of correct answers is then calculated as the ratio of the number of correct answers
        to the total number of answers.

        Overall, this function evaluates the quality of the answers provided, based on the specified
        minimal and maximal scores, and the proportion of correct answers.
        """
        count_ok = len(answers.checked & answers.correct) + len(answers.unchecked & answers.incorrect)
        proportion = count_ok / len(answers.all)
        return round(score.incorrect + proportion * (score.correct - score.incorrect), 2)

    @staticmethod
    def partial_answers(answers: AnswersData, score: ScoreData) -> float:
        """Return 0 if an incorrect answer was checked,
        else the proportion of correct answers."""
        return _partial_answers(answers, score)

    @staticmethod
    def partial_answers_quadratic(answers: AnswersData, score: ScoreData) -> float:
        """Return 0 if an incorrect answer was checked,
        else the squared proportion of correct answers."""
        return _partial_answers(answers, score, exposant=2)

    @staticmethod
    def correct_minus_incorrect(answers: AnswersData, score: ScoreData) -> float:
        """If the proportion of correct answers is less than 0.5, as is typically the case,
        return the difference between the proportion of correctly checked answers
        and the proportion of incorrectly checked answers.

        Otherwise, if the proportion of correct answers is greater than 0.5,
        return the difference between the proportion of correctly unchecked answers
        and the proportion of incorrectly unchecked answers.

        The idea behind this algorithm is that, if the proportion of correct answers
        is greater than 0.5, the exercise's difficulty lies in determining
        which answers should be left unchecked.

        Note that the
        """
        return _correct_minus_incorrect(answers, score)

    @staticmethod
    def correct_minus_incorrect_quadratic(answers: AnswersData, score: ScoreData) -> float:
        """Same algorithm as for `correct_minus_incorrect`, but the result is squared.

        This makes very unlikely for a student answering randomly to gain significant score.
        """
        return _correct_minus_incorrect(answers, score, exponent=2)


def _checked_among_correct_proportion(answers: AnswersData) -> float:
    """Return the proportion of checked answers among correct ones.

    If there was no correct answer, return 1.0 if no answer was checked, 0.0 else.
    """
    if len(answers.correct) == 0:
        return float(len(answers.checked) == 0)
    return len(answers.checked & answers.correct) / len(answers.correct)


def _partial_answers(answers: AnswersData, score: ScoreData, exposant=1.0) -> float:
    """Return 0 if an incorrect answer was checked, else the proportion of correct answers."""
    if answers.checked & answers.incorrect:
        return score.incorrect
    else:
        return round(_checked_among_correct_proportion(answers) ** exposant * score.correct, 2)


def _correct_minus_incorrect(answers: AnswersData, score: ScoreData, exponent=1.0) -> float:
    """If the proportion of correct answers is less than 0.5, as is typically the case,
    return the difference between the proportion of correctly checked answers
    and the proportion of incorrectly checked answers.

    Otherwise, if the proportion of correct answers is greater than 0.5,
    return the difference between the proportion of correctly unchecked answers
    and the proportion of incorrectly unchecked answers.

    The idea behind this algorithm is that, if the proportion of correct answers
    is greater than 0.5, the exercise's difficulty lies in determining
    which answers should be left unchecked.

    We then raise the result to the power of `exponent`.
    """
    ratio = 1 - len(answers.incorrectly_answered) / max(1, min(len(answers.correct), len(answers.incorrect)))
    return max(0.0, round(ratio**exponent * score.correct, 2))



import csv
from hashlib import blake2b
from multiprocessing import Pool
from pathlib import Path
from shutil import rmtree
from typing import Iterator

from PIL import Image
from numpy import ndarray, array, int8

from ptyx_mcq.scan.document_data import DocumentData, PicData, DetectionStatus, RevisionStatus
from ptyx_mcq.scan.paths_handler import PathsHandler, DirsPaths, FilesPaths
from ptyx_mcq.scan.pdftools import number_of_pages, extract_pdf_pictures, PIC_EXTS
from ptyx_mcq.tools.config_parser import (
    Configuration,
    get_answers_with_status,
    DocumentId,
    StudentName,
    StudentId,
    OriginalQuestionAnswersDict,
)
from ptyx_mcq.tools.extend_literal_eval import extended_literal_eval


def pic_names_iterator(data: dict[DocumentId, DocumentData]) -> Iterator[Path]:
    """Iterate over all pics found in data (i.e. all the pictures already analysed)."""
    for doc_data in data.values():
        for pic_data in doc_data["pages"].values():
            path = Path(pic_data.pic_path)
            # return pdfhash/picnumber.png
            yield path.relative_to(path.parent.parent)


class DataStorage:
    """Store and retrieve the data."""

    def __init__(self, config_path: Path, input_dir: Path = None, output_dir: Path = None):
        self.paths = PathsHandler(config_path=config_path, input_dir=input_dir, output_dir=output_dir)
        # All data extracted from pdf files.
        self.data: dict[DocumentId, DocumentData] = {}
        # Additional information entered manually.
        self.more_infos: dict[DocumentId, tuple[StudentName, StudentId]] = {}
        # Manually verified pages.
        self.verified: set[Path] = set()
        self.skipped: set[Path] = set()
        self.correct_answers: dict[DocumentId, OriginalQuestionAnswersDict] = {}
        self.neutralized_answers: dict[DocumentId, OriginalQuestionAnswersDict] = {}
        # self.paths.make_dirs()
        self.config: Configuration = self.get_configuration(self.paths.configfile)

    # def student_name_to_doc_id_dict(self) -> dict[StudentName, DocumentId]:
    #     """`student_name_to_doc_id` is used to retrieve the data associated with a name."""
    #     return {doc_data["name"]: doc_id for doc_id, doc_data in self.data.items()}

    @property
    def dirs(self) -> DirsPaths:
        return self.paths.dirs

    @property
    def files(self) -> FilesPaths:
        return self.paths.files

    @property
    def input_dir(self):
        return self.paths.input_dir

    @property
    def output_dir(self):
        return self.paths.output_dir

    def write_log(self, msg: str) -> None:
        with open(self.paths.logfile_path, "a", encoding="utf8") as logfile:
            logfile.write(msg)

    def get_pic(self, doc_id: int, page: int) -> Image.Image:
        webp = next(self.dirs.data.glob(f"{doc_id}-{page}.webp"))
        return Image.open(str(webp))

    def get_matrix(self, doc_id: int, page: int) -> ndarray:
        # noinspection PyTypeChecker
        return array(self.get_pic(doc_id, page).convert("L")) / 255

    def get_pics_list(self):
        """Return sorted pics list."""
        return sorted(f for f in self.dirs.pic.glob("*/*") if f.suffix.lower() in PIC_EXTS)

    def relative_pic_path(self, pic_path: str | Path):
        return Path(pic_path).relative_to(self.dirs.pic)

    def absolute_pic_path(self, pic_path: str | Path):
        return self.dirs.pic / pic_path

    def get_configuration(self, path: Path) -> Configuration:
        """Read configuration file, load configuration and calculate maximal score too."""
        cfg: Configuration = Configuration.load(path)
        self.correct_answers = get_answers_with_status(cfg, correct=True)
        self.neutralized_answers = get_answers_with_status(cfg, correct=None)
        default_mode = cfg.mode["default"]
        default_correct = cfg.correct["default"]

        max_score: float = 0
        # Take a random student test, and calculate max score for it.
        # Maximal score = (number of questions)x(score when answer is correct)
        for q in next(iter(cfg.ordering.values()))["questions"]:
            if cfg.mode.get(q, default_mode) != "skip":
                max_score += cfg.correct.get(q, default_correct)
        cfg.max_score = max_score
        return cfg

    def reload(self, reset=False) -> None:
        """Load all information from files."""
        self.paths.make_dirs(reset)
        self._update_input_data()
        # Load data from previous run.
        self._load_data()
        self._load_more_infos()
        self._load_manually_verified_pages_list()
        self._load_skipped_pictures_list()

    def _load_data(self) -> None:
        if self.dirs.data.is_dir():
            for filename in self.dirs.data.glob("*.scandata"):
                print(f"Loading: {filename}")
                doc_id = DocumentId(int(filename.stem))
                try:
                    with open(filename) as f:
                        self.data[doc_id] = extended_literal_eval(
                            f.read(),
                            {
                                "PicData": PicData,
                                "CHECKED": DetectionStatus.CHECKED,
                                "UNCHECKED": DetectionStatus.UNCHECKED,
                                "PROBABLY_CHECKED": DetectionStatus.PROBABLY_CHECKED,
                                "PROBABLY_UNCHECKED": DetectionStatus.PROBABLY_UNCHECKED,
                                "MARKED_AS_CHECKED": RevisionStatus.MARKED_AS_CHECKED,
                                "MARKED_AS_UNCHECKED": RevisionStatus.MARKED_AS_UNCHECKED,
                            },
                        )
                except Exception:
                    print(f"ERROR when reading {filename} :")
                    raise

    def _load_more_infos(self) -> None:
        """Retrieve manually entered information (if any)."""
        if self.files.more_infos.is_file():
            with open(self.files.more_infos, "r", newline="") as csvfile:
                reader = csv.reader(csvfile)
                for row in reader:
                    try:
                        doc_id, name, student_id = row
                    except ValueError:
                        doc_id, name = row
                        student_id = ""
                    self.more_infos[DocumentId(int(doc_id))] = (StudentName(name), StudentId(student_id))
                print("Retrieved infos:", self.more_infos)

    def _load_manually_verified_pages_list(self):
        """Load the list of manually verified pages. They should not be verified anymore."""
        if self.files.verified.is_file():
            with open(self.files.verified, "r", encoding="utf8", newline="") as file:
                self.verified = set(Path(line.strip()) for line in file.readlines())
                print("Pages manually verified:")
                for path in self.verified:
                    print(f"    • {path}")

    def _load_skipped_pictures_list(self) -> None:
        """List skipped pictures. Next time, they will be skipped with no warning."""
        self.skipped = set(pic_names_iterator(self.data))
        if self.files.skipped.is_file():
            with open(self.files.skipped, "r", encoding="utf8", newline="") as file:
                self.skipped |= set(Path(line.strip()) for line in file.readlines())
                print("Pictures skipped:")
                for path in sorted(self.skipped):
                    print(f"    • {path}")

    def store_doc_data(self, pdf_hash: str, doc_id: DocumentId, p: int, matrix: ndarray = None) -> None:
        with open(self.dirs.data / f"{pdf_hash}.index", "a") as f:
            f.write(str(doc_id) + "\n")
        with open(self.dirs.data / f"{doc_id}.scandata", "w") as f:
            f.write(repr(self.data[doc_id]))
        # We will store a compressed version of the matrix.
        # (It would consume too much memory else).
        if matrix is not None:
            webp = self.dirs.data / f"{doc_id}-{p}.webp"
            Image.fromarray((255 * matrix).astype(int8)).save(str(webp), format="WEBP")

    def store_additional_info(self, doc_id: int, name: str, student_ID: str) -> None:
        with open(self.files.more_infos, "a", newline="") as csvfile:
            writerow = csv.writer(csvfile).writerow
            writerow([str(doc_id), name, student_ID])

    def store_skipped_pic(self, skipped_pic: Path) -> None:
        with open(self.files.skipped, "a", newline="", encoding="utf8") as file:
            file.write(f"{skipped_pic}\n")

    def store_verified_pic(self, verified_pic: Path) -> None:
        with open(self.files.verified, "a", newline="", encoding="utf8") as file:
            file.write(f"{verified_pic}\n")

    def _generate_current_pdf_hashes(self) -> dict[str, Path]:
        """Return the hashes of all the pdf files found in `scan/` directory.

        Return: {hash: pdf path}
        """
        hashes = dict()
        for path in self.paths.input_dir.glob("**/*.pdf"):
            with open(path, "rb") as pdf_file:
                hashes[blake2b(pdf_file.read(), digest_size=20).hexdigest()] = path
        return hashes

    def _update_input_data(self) -> None:
        """Test if input data has changed, and update it if needed."""
        hash2pdf: dict[str, Path] = self._generate_current_pdf_hashes()

        self._remove_obsolete_files(hash2pdf)

        # For each new pdf files, extract all pictures
        to_extract: list[tuple[Path, Path]] = []
        with Pool() as pool:
            for pdfhash, pdfpath in hash2pdf.items():
                folder = self.dirs.pic / pdfhash
                if not folder.is_dir():
                    to_extract.append((pdfpath, folder))
                elif number_of_pages(pdfpath) != len(
                    [f for f in folder.iterdir() if f.suffix.lower() in PIC_EXTS]
                ):
                    # Extraction was probably interrupted
                    rmtree(folder)
                    folder.mkdir()
                    to_extract.append((pdfpath, folder))
            pool.starmap(extract_pdf_pictures, to_extract)

    def _remove_obsolete_files(self, hash2pdf):
        """For each removed pdf files, remove corresponding pictures and data."""
        for path in self.dirs.pic.iterdir():
            if not path.is_dir():
                raise RuntimeError(
                    f'Folder "{path.parent}" should only contain folders.\n'
                    "You may clean it manually, or remove it with following command:\n"
                    f'rm -r "{path.parent}"'
                )
            if path.name not in hash2pdf:
                rmtree(path)
        for path in self.dirs.data.glob("*.index"):
            if path.stem not in hash2pdf:
                with open(path) as f:
                    doc_ids = set(f.read().split())
                for doc_id in doc_ids:
                    (self.dirs.data / f"{doc_id}.scandata").unlink(missing_ok=True)
                    for webp in self.dirs.data.glob(f"{doc_id}-*.webp"):
                        webp.unlink()
                path.unlink()



from dataclasses import dataclass
from pathlib import Path
from shutil import rmtree
from time import strftime

from ptyx_mcq.tools.io_tools import get_file_or_sysexit


@dataclass
class FilesPaths:
    verified: Path
    skipped: Path
    more_infos: Path
    csv_scores: Path
    xlsx_scores: Path
    infos: Path


@dataclass
class DirsPaths:
    root: Path
    data: Path
    cfg: Path
    pic: Path
    log: Path
    pdf: Path


class PathsHandler:
    """Handles the different files and directories used by the scanner.

    The most import file for scanning is the configuration file, with extension ".ptyx.mcq.config.json".
    """

    configfile: Path
    input_dir: Path
    output_dir: Path

    def __init__(self, config_path: Path, input_dir: Path = None, output_dir: Path = None):
        self.configfile = get_file_or_sysexit(config_path, extension=".ptyx.mcq.config.json")

        root = self.configfile.parent

        if input_dir is None:
            input_dir = root / "scan"
        self.input_dir = input_dir

        if output_dir is None:
            output_dir = root / ".scan"
        self.output_dir = output_dir
        # `.scan` directory is used to write intermediate files.
        # Directory tree:
        # .scan/
        # .scan/pic -> pictures extracted from the pdf
        # .scan/cfg/more_infos.csv -> missing students names.
        # .scan/cfg/verified.csv -> pages already verified.
        # .scan/cfg/skipped.csv -> pages to skip.
        # .scan/scores.csv
        # .scan/data -> data stored as .scandata files (used to resume interrupted scan).

        cfg = output_dir / "cfg"
        log = output_dir / "log"
        self.dirs = DirsPaths(
            root=root,
            cfg=cfg,
            data=output_dir / "data",
            pic=output_dir / "pic",
            pdf=output_dir / "pdf",
            log=log,
        )
        self.files = FilesPaths(
            verified=cfg / "verified.txt",
            skipped=cfg / "skipped.txt",
            more_infos=cfg / "more_infos.csv",
            csv_scores=output_dir / "scores.csv",
            xlsx_scores=output_dir / "scores.xlsx",
            infos=output_dir / "infos.csv",
        )
        self.logfile_path = log / (strftime("%Y.%m.%d-%H.%M.%S") + ".log")

    def make_dirs(self, reset=False):
        """Make output directory structure. Is reset is `True`, remove all output directory content."""
        if reset and self.output_dir.is_dir():
            rmtree(self.output_dir)
        # noinspection PyUnresolvedReferences
        for dirname in self.dirs.__dataclass_fields__:
            getattr(self.dirs, dirname).mkdir(parents=True, exist_ok=True)



import sys
import traceback
from pathlib import Path

ANSI_RESET = "\u001B[0m"
ANSI_BLACK = "\u001B[30m"
ANSI_RED = "\u001B[31m"
ANSI_GREEN = "\u001B[32m"
ANSI_YELLOW = "\u001B[33m"
ANSI_BLUE = "\u001B[34m"
ANSI_PURPLE = "\u001B[35m"
ANSI_CYAN = "\u001B[1;36m"
ANSI_WHITE = "\u001B[37m"
ANSI_GRAY = "\u001B[90m"
ANSI_BOLD = "\u001B[1m"
ANSI_REVERSE_PURPLE = "\u001B[45m"
ANSI_REVERSE_BLUE = "\u001B[44m"


shell_colors = {
    "red": 1,
    "green": 2,
    "yellow": 3,
    "blue": 4,
}


def custom_print(msg: str, color: str, label: str, bold=False, **kw) -> None:
    n = shell_colors[color]
    print(f"\33[3{n}m[\33[9{n}{';1' if bold else ''}m{label}\33[0m\33[3{n}m]\33[0m " + msg, **kw)


def print_error(msg: str) -> None:
    custom_print(msg, "red", "Error", bold=True, file=sys.stderr)


def print_warning(msg: str) -> None:
    custom_print(msg, "yellow", "Warning", bold=False)


def print_success(msg: str) -> None:
    custom_print(msg, "green", "OK", bold=True)


def print_info(msg: str) -> None:
    custom_print(msg, "blue", "Info", bold=False)


def get_file_or_sysexit(path: Path, *, extension: str) -> Path:
    """Get the path of the ptyx file corresponding to the given `path`.

    If the file is not found, exit calling `sys.exit(1)`.
    """
    try:
        return get_file_with_extension(path, extension=extension)
    except OSError:
        traceback.print_exc()
        resolve = f" ({path.resolve()})" if str(path) != path.resolve() else ""
        print(f"Searching for a '{extension}' file in '{path}'{resolve} failed !")
        print_error(f"No '{extension}' file found.")
        sys.exit(1)


def get_file_with_extension(path: Path, *, extension) -> Path:
    """Get the path of the ptyx file corresponding to the given `path`.

    If `path` is already the path of a ptyx file, just return `path` unchanged.
    Else, `path` must be a directory which contains a single ptyx file, and the path
    of this ptyx file is returned.

    Raise `FileNotFoundError` if `path` is neither a directory nor a ptyx file, or if `path`
    is a directory but contains no ptyx file or several ptyx files.
    """
    path = path.expanduser().resolve()
    if path.suffix == extension:
        if not path.is_file():
            raise FileNotFoundError(f"File '{path}' not found.")
    else:
        found_files: list[Path] = list(path.glob("*" + extension))
        if len(found_files) == 0:
            raise FileNotFoundError(f"No '{extension}' file found in '{path}'.")
        elif len(found_files) > 1:
            raise FileNotFoundError(
                f"Several {extension} file found in '{path}', I don't know which one to chose."
            )
        path = found_files[0]
    return path


# config_path = config_path.expanduser().resolve()
#         if config_path.is_dir():
#             self.configfile = search_by_extension(config_path, ".ptyx.mcq.config.json")
#         elif config_path.is_file():
#             self.configfile = config_path.with_suffix(".ptyx.mcq.config.json")
#             if not self.configfile.is_file():
#                 raise FileNotFoundError(f"File '{self.configfile}' not found.")


# def search_by_extension(directory: Path, ext: str) -> Path:
#     """Search for a file with extension `ext` in given directory.
#
#     Search is NOT case sensible.
#     If no or more than one file is found, an error is raised.
#     """
#     ext = ext.lower()
#     paths = [pth for pth in directory.glob("*") if pth.name.lower().endswith(ext)]
#     if not paths:
#         raise FileNotFoundError(f"No {ext!r} file found in that directory: {directory} ! ")
#     elif len(paths) > 1:
#         raise RuntimeError(
#             f"Several {ext!r} file found in that directory: {directory} ! "
#             "Keep only one of them and delete all the others (or rename their extensions)."
#         )
#     return paths[0]
def print_framed_msg(msg: str) -> None:
    decoration = max(len(line) for line in msg.split("\n")) * "-"
    print(decoration)
    print(msg)
    print(decoration)



import csv
import math
from typing import TYPE_CHECKING, Callable

from ptyx_mcq.scan.evaluation_strategies import AnswersData, ScoreData, EvaluationStrategies
from ptyx_mcq.tools.io_tools import (
    print_info,
    ANSI_RESET,
    ANSI_RED,
    ANSI_GREEN,
    ANSI_YELLOW,
    ANSI_CYAN,
    ANSI_BOLD,
)

if TYPE_CHECKING:
    from ptyx_mcq.scan.scan import MCQPictureParser


class ScoresManager:
    def __init__(self, mcq_parser: "MCQPictureParser"):
        self.mcq_parser = mcq_parser
        self.scores: dict[str, float | str] = {}
        self.results: dict[str, float] = {}

    @property
    def max_score(self):
        return self.mcq_parser.config.max_score

    def calculate_scores(self) -> None:
        cfg = self.mcq_parser.config
        default_mode = cfg.mode["default"]
        default_weight = cfg.weight["default"]
        default_correct = cfg.correct["default"]
        default_incorrect = cfg.incorrect["default"]
        default_skipped = cfg.skipped["default"]
        default_floor = cfg.floor["default"]
        default_ceil = cfg.ceil["default"]

        data_manager = self.mcq_parser.data_storage
        for doc_id in self.mcq_parser.data:
            correct_ans = data_manager.correct_answers[doc_id]
            neutralized_ans = data_manager.neutralized_answers[doc_id]
            doc_data = self.mcq_parser.data[doc_id]
            print(f'Test {doc_id} - {doc_data["name"]}')
            for q in sorted(doc_data["answered"]):
                answered = set(doc_data["answered"][q])
                correct_ones = correct_ans[q]
                neutralized_ones = neutralized_ans[q]
                all_answers = {ans_num for ans_num, is_ok in cfg.ordering[doc_id]["answers"][q]}

                # Neutralized answers must be removed from each set of answers.
                # (Typically, neutralized answers are answers which were detected faulty during or after the
                # examination).
                answered -= neutralized_ones
                correct_ones -= neutralized_ones
                all_answers -= neutralized_ones

                mode = cfg.mode.get(q, default_mode)

                if mode == "skip":
                    # Used mostly to skip bogus questions.
                    print(f"Question {q} skipped...")
                    continue

                try:
                    func = getattr(EvaluationStrategies, mode)
                except AttributeError:
                    raise AttributeError(f"Unknown evaluation mode: {mode!r}.")

                ans_data = AnswersData(checked=answered, correct=correct_ones, all=all_answers)
                scores_data = ScoreData(
                    correct=float(cfg.correct.get(q, default_correct)),
                    skipped=float(cfg.skipped.get(q, default_skipped)),
                    incorrect=float(cfg.incorrect.get(q, default_incorrect)),
                )
                earn = func(ans_data, scores_data)

                floor = cfg.floor.get(q, default_floor)
                assert floor is None or isinstance(floor, (float, int))
                if floor is not None and earn < floor:
                    earn = floor
                ceil = cfg.ceil.get(q, default_ceil)
                assert ceil is None or isinstance(ceil, (float, int))
                if ceil is not None and earn > ceil:
                    earn = ceil

                if earn == scores_data.correct:
                    color = ANSI_GREEN
                elif earn == scores_data.incorrect:
                    color = ANSI_RED
                else:
                    color = ANSI_YELLOW
                print(f"-  {color}Rating (Q{q}): {color}{earn:g}{ANSI_RESET}")
                # Don't forget to include the weight of the question to calculate the global score.
                earn *= float(cfg.weight.get(q, default_weight))
                # Don't use weight for per question score, since it would make success rates
                # harder to compare.
                doc_data["score_per_question"][q] = earn
                doc_data["score"] += earn
        default = self.mcq_parser.config.default_score
        self.scores = {name: default for name in self.mcq_parser.config.students_ids.values()}
        self.results = {doc_data["name"]: doc_data["score"] for doc_data in self.mcq_parser.data.values()}
        self.scores.update(self.results)

    @staticmethod
    def _convert(num: float | str, factor=1.0):
        if isinstance(num, float):
            num = round(factor * num, 2)
        return num

    def print_scores(self) -> None:
        min_score: float = math.inf
        max_score: float = -math.inf
        print()
        print(f"{ANSI_CYAN}SCORES (/{self.max_score:g}):{ANSI_RESET}")
        for name, score in sorted(self.scores.items()):
            score = self._convert(score)
            if isinstance(score, (float, int)):
                if score < min_score:
                    min_score = score
                if score > max_score:
                    max_score = score
            print(f" - {name}: {self._convert(score)}")
        if self.results:
            mean = round(sum(self.results.values()) / len(self.results), 2)
            print(
                f"{ANSI_YELLOW}Mean: "
                f"{ANSI_BOLD}{mean:g}{ANSI_RESET}{ANSI_YELLOW}/{self.max_score:g}{ANSI_RESET}"
            )
            print(f"Min: {ANSI_RED}{min_score}{ANSI_RESET} - Max: {ANSI_GREEN}{max_score}{ANSI_RESET}")
        else:
            print("No score found !")
        print()

    def _write_scores(self, writerow: Callable) -> None:
        max_score = self.max_score
        writerow(("Name", f"Score/{max_score:g}", "Score/20", "Score/100"))
        for name, score in sorted(self.scores.items()):
            # TODO: Add ability to change the notation system.
            writerow(
                [
                    name,
                    self._convert(score),
                    self._convert(score, factor=20 / max_score) if max_score else 0,
                    self._convert(score, factor=100 / max_score) if max_score else 0,
                ]
            )

    def generate_csv_file(self) -> None:
        scores_path = self.mcq_parser.data_storage.files.csv_scores
        with open(scores_path, "w", newline="") as csvfile:
            self._write_scores(writerow=csv.writer(csvfile).writerow)
        print_info(f'Results stored in "{scores_path}"')

    def generate_xlsx_file(self) -> None:
        from openpyxl import Workbook
        from openpyxl.worksheet.worksheet import Worksheet
        from openpyxl.worksheet.table import Table, TableStyleInfo
        from openpyxl.utils import get_column_letter

        wb = Workbook()
        # grab the active worksheet
        sheet: Worksheet = wb.active  # type: ignore
        sheet.title = "Resume"
        self._write_scores(writerow=sheet.append)
        tab = Table(displayName="Table1", ref=f"A1:{get_column_letter(sheet.max_column)}{sheet.max_row}")

        # Add a default style with striped rows and banded columns
        style = TableStyleInfo(
            name="TableStyleMedium9",
            showFirstColumn=False,
            showLastColumn=False,
            showRowStripes=True,
            showColumnStripes=True,
        )
        tab.tableStyleInfo = style
        sheet.column_dimensions["A"].width = 1.23 * max(len(name) for name in self.scores)
        wb.save(self.mcq_parser.data_storage.files.xlsx_scores)



import string
import subprocess
from pathlib import Path

from ptyx_mcq.scan.color import Color, RGB
from ptyx_mcq.scan.data_manager import DataStorage
from ptyx_mcq.scan.document_data import Page, DetectionStatus, RevisionStatus
from ptyx_mcq.scan.visual_debugging import ArrayViewer
from ptyx_mcq.tools.config_parser import (
    DocumentId,
    ApparentQuestionNumber,
    apparent2real,
    ApparentAnswerNumber,
    StudentName,
    StudentId,
)
from ptyx_mcq.tools.io_tools import print_framed_msg, print_warning


class ConflictSolver:
    def __init__(self, data_storage: DataStorage):
        self.data_storage = data_storage
        self.data = data_storage.data

    def resolve_conflicts(self):
        """Resolve conflicts manually: unknown student ID, ambiguous answer..."""
        print("Searching for missing names...")
        self.review_missing_names()
        print("Searching for duplicate names...")
        self.review_duplicate_names()
        print("Searching for ambiguous answers...")
        self.review_answers()

    def review_missing_names(self):
        # First, complete missing information with previous scan data, if any.
        for doc_id, (student_name, student_id) in self.data_storage.more_infos.items():
            self.data[doc_id]["name"] = student_name
            self.data[doc_id]["student_ID"] = student_id
        # Search for any missing information remaining.
        for doc_id, doc_data in self.data.items():
            if doc_data["name"] == "":
                print_warning(f"No student name for document {doc_id}.")
                student_name, student_id = self.enter_name_and_id(doc_id)
                doc_data["name"] = student_name
                doc_data["student_ID"] = student_id
                self.data_storage.more_infos[doc_id] = (student_name, student_id)

    def review_duplicate_names(self) -> None:
        name_to_doc_id: dict[StudentName, DocumentId] = {}
        for doc_id, doc_data in self.data.items():
            name = doc_data["name"]
            if name in name_to_doc_id:
                matching_doc_id = name_to_doc_id[name]
                matching_doc_data = self.data[matching_doc_id]
                print_warning(f"Same student name on documents {doc_id} and {matching_doc_id}: {name!r}.")
                self.resolve_duplicate_name_conflict(name, doc_id, name_to_doc_id)
                self.data_storage.more_infos[doc_id] = doc_data["name"], doc_data["student_ID"]
                self.data_storage.more_infos[matching_doc_id] = (
                    matching_doc_data["name"],
                    matching_doc_data["student_ID"],
                )

    def resolve_duplicate_name_conflict(
        self, name: StudentName, doc_id: DocumentId, name_to_doc_id: dict[StudentName, DocumentId]
    ) -> None:
        while name in name_to_doc_id:
            print(f"Test #{name_to_doc_id[name]}: {name}")
            print(f"Test #{doc_id}: {name}")
            print_framed_msg(
                f"Error : 2 tests for same student ({name}) !\n"
                "Please modify at least one name (enter nothing to keep a name)."
            )
            # Remove twin name from name2doc_id, and get the corresponding previous test ID.
            doc_id0 = name_to_doc_id.pop(name)
            # Ask for a new name.
            name0, student_id0 = self.enter_name_and_id(doc_id0, default=name)
            # Update all infos.
            name_to_doc_id[name0] = doc_id0
            self.data[doc_id0]["name"] = name0
            self.data[doc_id0]["student_ID"] = student_id0
            # Ask for a new name for new test too.
            name = self.enter_name_and_id(doc_id)[0]

        assert name, "Name should not be empty at this stage !"
        name_to_doc_id[name] = doc_id
        self.data[doc_id]["name"] = name

    def enter_name_and_id(self, doc_id: DocumentId, default: str = "") -> tuple[StudentName, StudentId]:
        array = self.data_storage.get_matrix(doc_id, Page(1))
        width = array.shape[1]
        viewer = ArrayViewer(array[0 : int(3 / 4 * width), :])
        student_ids = self.data_storage.config.students_ids
        student_ID = ""
        print("Name can not be read automatically.")
        print("Please read the name on the picture which will be displayed now.")
        input("-- Press enter --")
        #    subprocess.run(["display", "-resize", "1920x1080", pic_path])
        # TODO: use first letters of students name to find student.
        # (ask again if not found, or if several names match first letters)
        process = None
        name = ""
        while name == "":
            # Don't relaunch process if it is still alive.
            # (process.poll() is not None for dead processes.)
            if process is None or process.poll() is not None:
                process = viewer.display(wait=False)
            name = input("Student name or ID:").strip()
            if not name:
                name = default
            elif student_ids:
                if name in student_ids:
                    name, student_ID = student_ids[StudentId(name)], name
                elif any((digit in name) for digit in string.digits):
                    # This is not a student name !
                    print("Unknown ID.")
                    name = ""
            if name:
                print("Name: %s" % name)
                if input("Is it correct ? (Y/n)").lower() not in ("y", "yes", ""):
                    name = ""
        assert process is not None
        process.terminate()
        # Keep track of manually entered information (will be useful if the scan has to be run again later !)
        self.data_storage.store_additional_info(doc_id=doc_id, name=name, student_ID=student_ID)
        return StudentName(name), StudentId(student_ID)

    def review_answers(self):
        for doc_id, doc_data in self.data.items():
            for page, pic_data in doc_data["pages"].items():
                if pic_data.needs_review and Path(pic_data.pic_path) not in self.data_storage.verified:
                    # The answers are ambiguous and were not already manually verified in a previous scan.
                    print_warning(f"Ambiguous answers for student {doc_data['name']}.")
                    self.edit_answers(doc_id, page)
                    self.data_storage.store_verified_pic(Path(pic_data.pic_path))

    def display_picture_with_detected_answers(self, doc_id: DocumentId, page: Page) -> subprocess.Popen:
        """Display the page with its checkboxes colored following their detection status."""
        array = self.data_storage.get_matrix(doc_id, page)
        viewer = ArrayViewer(array)
        pic_data = self.data[doc_id]["pages"][page]
        colors: dict[DetectionStatus | RevisionStatus, RGB] = {
            DetectionStatus.CHECKED: Color.blue,
            DetectionStatus.PROBABLY_CHECKED: Color.green,
            DetectionStatus.PROBABLY_UNCHECKED: Color.magenta,
            DetectionStatus.UNCHECKED: Color.pink,
            RevisionStatus.MARKED_AS_CHECKED: Color.cyan,
            RevisionStatus.MARKED_AS_UNCHECKED: Color.red,
        }
        thicknesses: dict[DetectionStatus | RevisionStatus, int] = {
            DetectionStatus.CHECKED: 2,
            DetectionStatus.PROBABLY_CHECKED: 5,
            DetectionStatus.PROBABLY_UNCHECKED: 5,
            DetectionStatus.UNCHECKED: 2,
            RevisionStatus.MARKED_AS_CHECKED: 5,
            RevisionStatus.MARKED_AS_UNCHECKED: 5,
        }
        for (q, a), (i, j) in pic_data.positions.items():
            status = pic_data.revision_status.get((q, a), pic_data.detection_status[(q, a)])
            viewer.add_square((i, j), pic_data.cell_size, color=colors[status], thickness=thicknesses[status])
        return viewer.display(wait=False)

    def edit_answers(self, doc_id: DocumentId, page: Page) -> None:
        """Interactive editor to change answers."""
        config = self.data_storage.config
        pic_data = self.data[doc_id]["pages"][page]
        # m = self.data_storage.get_matrix(doc_id, page)
        # cell_size = pic_data.cell_size
        answered = pic_data.answered
        revision_status = pic_data.revision_status
        print("Please verify answers detection:")
        input("-- Press ENTER --")

        while True:
            process = self.display_picture_with_detected_answers(doc_id, page)
            if input("Is this correct ? [(y)es/(N)o]").lower() in ("y", "yes"):
                break
            while (q_str := input("Write a question number, or 0 to escape:")) != "0":
                try:
                    q0 = ApparentQuestionNumber(int(q_str))
                    q, _ = apparent2real(q0, None, config, doc_id)
                    if q not in answered:
                        raise IndexError(rf"Invalid question number: {q0}")

                    checked = answered[q]
                    a_str = input(
                        "Add or remove answers "
                        "(Example: +2 -1 -4 to add answer 2, and remove answers 1 et 4):"
                    )
                    for val in a_str.split():
                        op, a0 = val[0], ApparentAnswerNumber(int(val[1:]))
                        q, a = apparent2real(q0, a0, config, doc_id)
                        if op == "+":
                            if a in checked:
                                print(f"Warning: {a0} already in answers.")
                            else:
                                checked.add(a)
                                revision_status[(q, a)] = RevisionStatus.MARKED_AS_CHECKED
                        elif op == "-":
                            if a in checked:
                                checked.remove(a)
                                revision_status[(q, a)] = RevisionStatus.MARKED_AS_UNCHECKED
                            else:
                                print(f"Warning: {a0} not in answers.")
                        else:
                            print(f"Invalid operation: {val!r}")
                except ValueError:
                    print("Invalid value.")
                    continue
                except (KeyError, IndexError):
                    print("Invalid number.")
                finally:
                    process.terminate()
                    process = self.display_picture_with_detected_answers(doc_id, page)
        process.terminate()

    def display_first_page_head(self, doc_id: DocumentId) -> subprocess.CompletedProcess | subprocess.Popen:
        """Display the head of the first page, to be able to read student name manually."""
        array = self.data_storage.get_matrix(doc_id, Page(1))
        width = array.shape[1]
        viewer = ArrayViewer(array[0 : int(3 / 4 * width), :])
        return viewer.display(wait=False)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
ptyx MCQ Command Line Interface

@author: Nicolas Pourcelot
"""

import shutil
import sys
from argparse import ArgumentParser
from os import unlink
from pathlib import Path
from typing import Optional

from platformdirs import PlatformDirs
from ptyx.latex_generator import compiler

from ptyx_mcq import IncludeParser
from ptyx_mcq.scan.evaluation_strategies import EvaluationStrategies
from .make.make import make, parse_ptyx_file
from .scan.scan import scan
from .tools.config_parser import Configuration
from .tools.io_tools import (
    print_success,
    print_error,
    get_file_or_sysexit,
    ANSI_RESET,
    ANSI_REVERSE_PURPLE,
    ANSI_REVERSE_BLUE,
    ANSI_BLUE,
)


def main(args: Optional[list] = None) -> None:
    """Main entry point, called whenever `mcq` command is executed."""
    parser = ArgumentParser(description="Generate and manage pdf MCQs.")
    subparsers = parser.add_subparsers()
    add_parser = subparsers.add_parser

    # create the parser for the "new" command
    new_parser = add_parser("new", help="Create an empty ptyx file.")
    new_parser.add_argument("path", nargs="?", metavar="PATH", type=Path, default="new-mcq")
    new_parser.add_argument(
        "--include",
        "-i",
        metavar="INCLUDE_PATH",
        type=Path,
        help="Include all files from this path in the generated .ptyx file.",
    )
    new_parser.add_argument(
        "--template",
        "-t",
        metavar="TEMPLATE_NAME",
        type=str,
        default="",
        help=(
            "Specify the name of the template to use.\nIf not specified, search for a template named "
            "'default' in the user config directory, or use the default template.\n"
            "One may force the use of the default template by writing 'original'."
        ),
    )
    new_parser.set_defaults(func=new)

    # create the parser for the "make" command
    make_parser = add_parser("make", help="Generate pdf file.")
    make_parser.add_argument("path", nargs="?", metavar="PATH", type=Path, default=".")
    make_parser.add_argument(
        "--num",
        "-n",
        metavar="N",
        type=int,
        default=1,
        help="Specify how many versions of the document must be generated.",
    )
    make_parser.add_argument(
        "--start",
        "-s",
        metavar="START",
        type=int,
        default=1,
        help="First document number (default=1).",
    )
    make_parser.add_argument("--quiet", "-q", action="store_true", help="Hide pdflatex output.")
    make_parser.add_argument(
        "--correction-only", "-co", action="store_true", help="Generate only correction for fast testing."
    )
    make_parser.set_defaults(func=make)

    # create the parser for the "scan" command
    scan_parser = add_parser("scan", help="Generate scores from scanned documents.")
    scan_parser.add_argument(
        "path",
        nargs="?",
        default=".",
        help=(
            "Path to a directory which must contain "
            "a .mcq.config file and a .scan.pdf file "
            "(alternatively, this path may point to any file in this folder)."
        ),
    )
    scan_parser.add_argument(
        "--reset",
        action="store_true",
        help="Delete all cached data." "The scanning process will restart from the beginning.",
    )
    scan_parser.add_argument(
        "--verify",
        "--manual-verification",
        choices=("always", "never", "auto"),
        default="auto",
        help="If set to `always`, then for each page scanned, display a picture of "
        "the interpretation by the detection algorithm, "
        "for manual verification.\n"
        "If set to `never`, always assume algorithm is right.\n"
        "Default is `auto`, i.e. only ask for manual verification "
        "in case of ambiguity.",
    )
    scan_parser.add_argument(
        "--ask-for-name",
        action="store_true",
        default=False,
        help="For each first page, display a picture of " "the top of the page and ask for the student name.",
    )
    scan_parser.set_defaults(func=scan)

    # create the parser for the "clear" command
    clear_parser = add_parser("clear", help="Remove every MCQ data but the ptyx file.")
    clear_parser.add_argument("path", nargs="?", metavar="PATH", type=Path, default=".")
    clear_parser.set_defaults(func=clear)

    # create the parser for the "update-config" command
    update_config_parser = add_parser("update-config", help="Update mcq configuration file.")
    update_config_parser.add_argument("path", nargs="?", metavar="PATH", type=Path, default=".")
    update_config_parser.set_defaults(func=update_config)

    # create the parser for the "update-include" command
    update_include_parser = add_parser("update-include", help="Update included files.")
    update_include_parser.add_argument("path", nargs="?", metavar="PATH", type=Path, default=".")
    update_include_parser.set_defaults(func=update_include)

    # create the parser for the "create_template" command
    create_template_parser = add_parser("create-template", help="Create a customisable user template.")
    create_template_parser.add_argument(
        "name",
        nargs="?",
        metavar="NAME",
        type=str,
        default="default",
        help="The template name must be a valid directory name.",
    )
    create_template_parser.set_defaults(func=create_template)

    # create the parser for the "strategies" command
    strategies_parser = add_parser("strategies", help="List available evaluation strategies.")
    strategies_parser.set_defaults(func=strategies)

    parsed_args = parser.parse_args(args)
    try:
        # Launch the function corresponding to the given subcommand.
        kwargs = vars(parsed_args)
        func = kwargs.pop("func")
    except KeyError:
        # No subcommand passed.
        parser.print_help()
        return
    func(**kwargs)


def new(path: Path, include: Path = None, template="") -> None:
    """Implement `mcq new` command."""
    # Select the template to use.
    # Default template:
    template_path = Path(__file__).resolve().parent / "templates/original"
    # Directory of the eventual user templates:
    user_templates_path = PlatformDirs().user_config_path / "ptyx-mcq/templates"
    if template == "":
        # Search for a default user-defined template.
        user_default_template_path = user_templates_path / "default"
        if user_default_template_path.is_dir():
            template_path = user_default_template_path
    elif template != "original":
        template_path = user_templates_path / template
    if not template_path.is_dir():
        print_error(f"I can't use template {template!r}: '{template_path}' directory not found.")
        sys.exit(1)

    # Create the new MCQ.
    if path.exists():
        print_error(f"Path {path} already exists.")
        sys.exit(1)
    else:
        print(f"Using template from '{template_path}'.")
        shutil.copytree(template_path, path)
        if include is not None:
            # No need to have a questions directory template,
            # since questions' files will be explicitly listed.
            shutil.rmtree(path / "questions", ignore_errors=True)
            # Edit the .ptyx file, to replace default code with the list of the questions' files.
            ptyx_path = get_file_or_sysexit(path, extension=".ptyx")
            new_lines = [f"-- ROOT: {include.resolve()}"]
            for include_path in include.glob("**/*.ex"):
                new_lines.append(f"-- {include_path.relative_to(include)}")
            lines = []
            with open(ptyx_path, encoding="utf8") as f:
                include_section = False
                for line in f:
                    line = line.rstrip("\n")
                    if line.startswith("<<<<"):
                        include_section = True
                        lines.append(line)
                        lines.extend(new_lines)
                    elif line.startswith(">>>>"):
                        include_section = False
                        lines.append(line)
                    elif not include_section:
                        lines.append(line)

            with open(ptyx_path, "w", encoding="utf8") as f:
                f.write("\n".join(lines) + "\n")
        print_success(f"A new MCQ was created at {path}.")


def clear(path: Path) -> None:
    """Implement `mcq clear` command."""
    ptyxfile_path = get_file_or_sysexit(path, extension=".ptyx")
    filename = ptyxfile_path.name
    root = ptyxfile_path.parent
    for directory in (".scan", ".compile"):
        try:
            shutil.rmtree(root / directory)
        except FileNotFoundError as e:
            print(f"Warning: {e}")
            print(f"Info: '{directory}' not found...")
    for filepath in (
        ptyxfile_path.with_suffix(".pdf"),
        ptyxfile_path.with_suffix(".ptyx.mcq.config.json"),
        ptyxfile_path.with_name(f".{filename}.plain-ptyx"),
        ptyxfile_path.with_name(f"{ptyxfile_path.stem}-corr.pdf"),
    ):
        try:
            unlink(filepath)
        except FileNotFoundError as e:
            print(f"Warning: {e}")
            print(f"Info: '{filepath}' not found...")
    print_success("Directory cleared.")


def update_config(path: Path) -> None:
    config_file = get_file_or_sysexit(path, extension=".ptyx.mcq.config.json")
    config = Configuration.load(config_file)
    parse_ptyx_file(path)
    # Compile ptyx file to LaTeX, to update information, but without compiling
    # LaTeX, which is slow (don't generate pdf files again).
    for doc_id in config.ordering:
        compiler.get_latex(PTYX_NUM=doc_id)
    # Update configuration
    data: Configuration = compiler.latex_generator.mcq_data
    # Test compatibility of `data` with `config`:
    # The number of questions and their ordering should not have change,
    # the same holds for the answers (but their labelling as correct
    # or incorrect is allowed to change).
    if not same_questions_and_answers_numbers(data, config):
        cmd = f"mcq make {path}"
        sep = len(cmd) * "-"
        cmd = f"{sep}\n{cmd}\n{sep}\n"
        print_error("Questions or answers changed.\n" "You must run a full compilation:\n" + cmd)

    # Get back the positions of checkboxes from last full compilation.
    data.id_table_pos = config.id_table_pos
    data.boxes = config.boxes
    data.dump(config_file.parent / (config_file.name.split(".")[0] + ".ptyx.mcq.config.json"))
    print_success("Configuration file was successfully updated.")


def same_questions_and_answers_numbers(config1: Configuration, config2: Configuration) -> bool:
    if list(config1.ordering) != list(config2.ordering):
        return False
    for doc_id in config1.ordering:
        ordering1 = config1.ordering[doc_id]
        ordering2 = config2.ordering[doc_id]
        if ordering1["questions"] != ordering2["questions"]:
            return False
        if list(ordering1["answers"]) != list(ordering2["answers"]):
            return False

        for q in ordering1["answers"]:
            if [a for a, _ in ordering1["answers"][q]] != [a for a, _ in ordering2["answers"][q]]:
                return False
    return True


def update_include(path: Path) -> None:
    """Update the list of included files."""
    ptyxfile_path = get_file_or_sysexit(path, extension=".ptyx")
    root = ptyxfile_path.parent
    IncludeParser(root).update(ptyxfile_path)
    print_success("The list of included files was successfully updated.")


def strategies() -> None:
    """Display all evaluation modes with a description."""
    strategies = EvaluationStrategies.get_all_strategies()
    print(f"\n{ANSI_REVERSE_PURPLE}[ Available strategies ]{ANSI_RESET}")
    print(", ".join(strategies))
    print()
    print(f"\n{ANSI_REVERSE_PURPLE}[ Details ]{ANSI_RESET}")
    for name in strategies:
        print(f"\n {ANSI_BLUE}╭───╴{ANSI_RESET}{ANSI_REVERSE_BLUE} {name} {ANSI_RESET} ")
        print(f" {ANSI_BLUE}│{ANSI_RESET} ")
        for line in getattr(EvaluationStrategies, name).__doc__.strip().split("\n"):
            print(f" {ANSI_BLUE}│{ANSI_RESET} " + line.strip())
        print(f" {ANSI_BLUE}│{ANSI_RESET} ")
        print(f" {ANSI_BLUE}╰───╴{ANSI_RESET}")


def create_template(name: str = "default") -> None:
    """Create default user template."""
    if name == "original":
        print_error(f"Name {name!r} is reserved, please choose another template name.")
        sys.exit(1)
    user_template = PlatformDirs().user_config_path / f"ptyx-mcq/templates/{name}"
    if user_template.is_dir():
        print_error(f"Folder {user_template} already exist.")
        sys.exit(1)
    default_template = Path(__file__).resolve().parent / "templates/original"
    shutil.copytree(default_template, user_template)
    print_success(f"Template created at {user_template}. Edit the inner files to customize it.")


if __name__ == "__main__":
    main()



from _ast import (
    Expression,
    Constant,
    UnaryOp,
    UAdd,
    USub,
    Tuple,
    List,
    Set,
    Name,
    Call,
    Dict,
    BinOp,
    Add,
    Sub,
)
from ast import parse


def extended_literal_eval(node_or_string, names: dict = None):
    """
    Safely evaluate an expression node or a string containing a Python
    expression.

    In python's `ast.literal_eval()`, the string or node provided
    may only consist of the following Python literal structures:
    strings, bytes, numbers, tuples, lists, dicts, sets, booleans, and None.

    This extends `ast.literal_eval()` by providing a `names` dictionary.
    To add support to `MyClass(...)`, you have to pass `names={"MyClass": MyClass}`.
    """
    if names is None:
        names = {}
    if isinstance(node_or_string, str):
        node_or_string = parse(node_or_string.lstrip(" \t"), mode="eval")
    if isinstance(node_or_string, Expression):
        node_or_string = node_or_string.body

    def _raise_malformed_node(node):
        msg = "malformed node or string"
        if lno := getattr(node, "lineno", None):
            msg += f" on line {lno}"
        raise ValueError(msg + f": {node!r}")

    def _convert_num(node):
        if not isinstance(node, Constant) or type(node.value) not in (int, float, complex):
            _raise_malformed_node(node)
        return node.value

    # noinspection PyTypeChecker
    def _convert_signed_num(node):
        if isinstance(node, UnaryOp) and isinstance(node.op, (UAdd, USub)):
            operand = _convert_num(node.operand)
            if isinstance(node.op, UAdd):
                return +operand
            else:
                return -operand
        return _convert_num(node)

    # noinspection PyTypeChecker
    def _convert(node):
        if isinstance(node, Constant):
            return node.value
        elif isinstance(node, Tuple):
            return tuple(map(_convert, node.elts))
        elif isinstance(node, List):
            return list(map(_convert, node.elts))
        elif isinstance(node, Set):
            return set(map(_convert, node.elts))
        elif isinstance(node, Name) and node.id in names:
            return names[node.id]
        elif (
            isinstance(node, Call)
            and isinstance(node.func, Name)
            and node.func.id == "set"
            and node.args == node.keywords == []
        ):
            return set()
        elif isinstance(node, Call) and isinstance(node.func, Name) and node.func.id in names:
            return names[node.func.id](
                *map(_convert, node.args), **{kw.arg: _convert(kw.value) for kw in node.keywords}
            )
        elif isinstance(node, Dict):
            if len(node.keys) != len(node.values):
                _raise_malformed_node(node)
            return dict(zip(map(_convert, node.keys), map(_convert, node.values)))
        elif isinstance(node, BinOp) and isinstance(node.op, (Add, Sub)):
            left = _convert_signed_num(node.left)
            right = _convert_num(node.right)
            if isinstance(left, (int, float)) and isinstance(right, complex):
                if isinstance(node.op, Add):
                    return left + right
                else:
                    return left - right
        return _convert_signed_num(node)

    return _convert(node_or_string)



import subprocess
import tempfile
from dataclasses import dataclass
from os.path import join
from typing import Any, Optional, overload, Literal

from PIL import Image
from numpy import ndarray, int8

from ptyx_mcq.scan.color import Color, RGB

FloatPosition = tuple[float, float]


# TODO: rewrite this using a class.


@dataclass
class RectangularShape:
    start: FloatPosition
    end: Optional[FloatPosition] = None
    color: RGB = Color.red
    thickness: int = 2
    fill: bool = False


class ArrayViewer:
    """`ArrayViewer` is used to display pictures annotated with colored rectangles, mainly for debugging.

    The picture to be displayed must be given as a numpy array.

    Parameters:
    `array` is an array containing the image data (image must be gray mode,
    each pixel represented by a float from 0 (black) to 1 (white)).

    Example:

        >> viewer = ArrayViewer(array)
        >> viewer.add_rectangle((2, 5), (10, 15), color=Color.red, thickness=4)
        >> viewer.add_rectangle((10, 10), (30, 20), color=Color.blue, fill=True)
        >> viewer.display()

    The annotated picture can be displayed using `.display()` methode.

    Use `.clear()` method to remove all annotations.

    NOTA:
    - `feh` must be installed.
      On Ubuntu/Debian: sudo apt-get install feh
    - Left-dragging the picture with mouse inside feh removes blurring/anti-aliasing,
      making visual debugging a lot easier.
    """

    _array: Optional[ndarray] = None
    _pic: Optional[Image.Image] = None

    def __init__(self, array: ndarray = None):
        self._shapes: list[RectangularShape] = []
        if array is not None:
            self.array = array

    @property
    def array(self) -> ndarray | None:
        return self._array

    @array.setter
    def array(self, array: ndarray) -> None:
        self._array = array
        self._shapes.clear()

    def add_area(
        self,
        start: FloatPosition,
        end: FloatPosition = None,
        color: RGB = Color.red,
        thickness: int = 2,
        fill: bool = False,
    ) -> None:
        """
        Annotate the image with a colored rectangle.

        Parameters:
            `start`: a couple of coordinates describing one corner of the rectangle.
            `end`: the opposite corner of the rectangle.
            `color`: the color of the rectangle (both the border and the interior if fill is true)
            `thickness`: the thickness of the border
            `fill`: if true, fill the rectangle.

        The color can be an attribute of the `Color` class (like `Color.red`), for common color's names,
        or a RGB tuple ([0-255], [0-255], [0-255]).

        """
        self._shapes.append(RectangularShape(start, end, color=color, thickness=thickness, fill=fill))

    def add_rectangle(
        self,
        start: FloatPosition,
        width: int,
        height: int,
        color: RGB = Color.red,
        thickness: int = 2,
        fill: bool = False,
    ) -> None:
        i, j = start
        self.add_area(start, (i + width, j + height), color=color, thickness=thickness, fill=fill)

    def add_square(
        self,
        start: FloatPosition,
        size: int,
        color: RGB = Color.red,
        thickness: int = 2,
        fill: bool = False,
    ) -> None:
        self.add_rectangle(start, size, size, color=color, thickness=thickness, fill=fill)

    def clear(self):
        """Clear all previous stored annotations."""
        self._shapes.clear()

    def _draw_rectangles(self) -> None:
        if self.array is not None:
            self._pic = Image.fromarray((255 * self.array).astype(int8)).convert("RGB")
            for rectangle in self._shapes:
                self._draw_rectangle(rectangle)

    def _draw_rectangle(self, rectangle: RectangularShape) -> None:
        if self.array is None:
            return
        height, width = self.array.shape
        i1, j1 = rectangle.start
        i2, j2 = rectangle.end or rectangle.start
        if i2 is None:
            i2 = height - 1
        if j2 is None:
            j2 = width - 1
        imin, imax = int(max(min(i1, i2), 0)), int(min(max(i1, i2), height - 1))
        jmin, jmax = int(max(min(j1, j2), 0)), int(min(max(j1, j2), width - 1))
        assert 0 <= imin <= imax < height and 0 <= jmin <= jmax < width

        color = rectangle.color
        thickness = rectangle.thickness
        assert self._pic is not None
        pix: Any = self._pic.load()  # type:ignore

        if rectangle.fill:
            for i in range(imin, imax + 1):
                for j in range(jmin, jmax + 1):
                    pix[j, i] = color
        else:
            # left and right sides of rectangle
            for i in range(imin, imax + 1):
                for j in range(jmin, jmin + thickness):
                    pix[j, i] = color
                for j in range(jmax + 1 - thickness, jmax + 1):
                    pix[j, i] = color
            # top and bottom sides of rectangle
            for j in range(jmin, jmax + 1):
                for i in range(imin, imin + thickness):
                    pix[j, i] = color
                for i in range(imax + 1 - thickness, imax + 1):
                    pix[j, i] = color

    @overload
    def display(self, wait: Literal[True] = True) -> subprocess.CompletedProcess:
        """Function signature when wait is True."""

    @overload
    def display(self, wait: Literal[False]) -> subprocess.Popen:
        """Function signature when wait is False."""

    def display(self, wait: bool = True) -> subprocess.CompletedProcess | subprocess.Popen:
        if self.array is None:
            raise RuntimeError("No picture to display.")
        self._draw_rectangles()
        if subprocess.call(["which", "feh"]) != 0:
            raise RuntimeError(
                "The `feh` command is not found, please " "install it (`sudo apt install feh` on Ubuntu)."
            )
        with tempfile.TemporaryDirectory() as tmpdirname:
            path = join(tmpdirname, "test.png")
            assert self._pic is not None
            self._pic.save(path)
            process: subprocess.CompletedProcess | subprocess.Popen
            if wait:
                process = subprocess.run(["feh", "-F", path])
            else:
                process = subprocess.Popen(["feh", "-F", path], stdin=subprocess.DEVNULL)
            input("-- pause --\n")
            return process


def color2debug(
    array: ndarray,
    from_: FloatPosition = None,
    to_: FloatPosition = None,
    color: RGB = Color.red,
    thickness: int = 2,
    fill=False,
    wait=True,
):
    """Display picture with a red (by default) rectangle for debugging.

    Parameters:
        `array`: an array containing the image data (image must be gray mode,
        each pixel represented by a float from 0 (black) to 1 (white)).
        `from_`: represent one corner of the red rectangle.
        `to_`: represent opposite corner of the red rectangle.
        `color`: color given as a RGB tuple ([0-255], [0-255], [0-255]).
        `thickness`: the thickness of the border
        `fill`: indicates if the rectangle should be filled.

    Usage: color2debug((0,0), (200,10), color=(255, 0, 255))

    NOTA:
    - `feh` must be installed.
      On Ubuntu/Debian: sudo apt-get install feh
    - Left-dragging the picture with mouse inside feh removes blurring/anti-aliasing,
      making visual debugging a lot easier.
    """
    viewer = ArrayViewer(array)
    if from_ is not None:
        viewer.add_area(from_, to_, color=color, thickness=thickness, fill=fill)
    viewer.display(wait=wait)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Aug 29 14:49:37 2019

@author: nicolas
"""
from os.path import join

# from typing import TYPE_CHECKING

from PIL import ImageDraw, ImageFont
from PIL.Image import Image
from ptyx_mcq.scan.data_manager import DataStorage
from .document_data import DocumentData

from .square_detection import Pixel
from .color import Color
from ..tools.config_parser import DocumentId, OriginalQuestionNumber, QuestionNumberOrDefault


# if TYPE_CHECKING:
#     from ptyx_mcq.scan.scanner import MCQPictureParser


def amend_all(data_storage: DataStorage) -> None:
    """Amend all generated documents, adding the scores and indicating the correct answers."""
    cfg = data_storage.config
    default_weight = cfg.weight["default"]
    assert isinstance(default_weight, (int, float))
    default_correct = cfg.correct["default"]
    assert isinstance(default_correct, (int, float))
    max_score_per_question: dict[QuestionNumberOrDefault, float] = {
        "default": default_weight * default_correct
    }
    for question in set(cfg.correct) | set(cfg.weight) - {"default"}:
        _max_score = cfg.weight.get(question, default_weight) * cfg.correct.get(question, default_correct)
        max_score_per_question[question] = _max_score

    # Global maximal score.
    # Note that the global maximal score can not be easily deduced from the dict `max_score_per_question`,
    # since this dict only include customized questions scores, and since many versions of the same question
    # may appear in the dict, but only one of those versions was included in each generated document.
    max_score = data_storage.config.max_score
    assert isinstance(max_score, (float, int)), repr(max_score)
    N = len(data_storage.data)
    for i, (doc_id, doc_data) in enumerate(data_storage.data.items(), start=1):
        print(f"Generating the amended pdf files: {i}/{N}...", end="\r")
        amend_doc(doc_data, doc_id, max_score, max_score_per_question, data_storage)
    print("Generating the amended pdf files: OK" + len(f"{N}/{N}...") * " ")


def amend_doc(
    doc_data: DocumentData,
    doc_id: DocumentId,
    max_score: float,
    max_score_per_question: dict[QuestionNumberOrDefault, float],
    data_storage: DataStorage,
) -> None:
    correct_answers = data_storage.correct_answers[doc_id]
    neutralized_answers = data_storage.neutralized_answers[doc_id]
    pics = {}
    for page, page_data in doc_data["pages"].items():
        top_left_positions: dict[OriginalQuestionNumber, Pixel] = {}
        # Convert to RGB picture.
        pic = data_storage.get_pic(doc_id, page).convert("RGB")
        if not page_data.positions:
            # The last page of the MCQ may be empty.
            # `float('+inf')` is used to ensure
            # it will be the last page when sorting.
            pics[float("+inf")] = pic
            continue
        # Drawing context
        draw = ImageDraw.Draw(pic)
        size = page_data.cell_size
        for (q, a), pos in page_data.positions.items():
            checked = a in page_data.answered[q]
            if a in neutralized_answers[q]:
                correct = None
            else:
                correct = a in correct_answers[q]
            _correct_checkboxes(draw, pos, checked, correct, size)
            if q in top_left_positions:
                i0, j0 = top_left_positions[q]
                i, j = pos
                top_left_positions[q] = (min(i, i0), min(j, j0))
            else:
                top_left_positions[q] = pos
        for q in top_left_positions:
            earn = doc_data["score_per_question"][q]
            maximum = max_score_per_question.get(q, max_score_per_question["default"])
            assert isinstance(maximum, (float, int)), repr(maximum)
            i, j = top_left_positions[q]
            _write_score(draw, (i, j - 2 * size), earn, maximum, size)
        # We will now sort pages.
        # For that, we use questions numbers: the page which displays
        # the smaller questions numbers is the first one, and so on.
        # However, be careful to use displayed questions numbers,
        # since `q` is the question number *before shuffling*.
        q_num = page_data.questions_nums_conversion[q]
        pics[q_num] = pic
        # Sort pages now.
    pages: list[Image]
    _, pages = zip(*sorted(pics.items()))  # type: ignore
    draw = ImageDraw.Draw(pages[0])
    _write_score(draw, (2 * size, 4 * size), doc_data["score"], max_score, 2 * size)
    pages[0].save(
        join(data_storage.dirs.pdf, f"{doc_data['name']}-{doc_id}.pdf"),
        save_all=True,
        append_images=pages[1:],
    )


def _correct_checkboxes(
    draw: ImageDraw.ImageDraw, pos: Pixel, checked: bool, correct: bool | None, size: int
) -> None:
    i, j = pos
    margin = size // 2
    # Draw a blue square around the box (for debugging purpose).
    draw.rectangle((j, i, j + size, i + size), outline=Color.green)
    red = Color.red
    if correct is None:
        draw.rectangle((j, i, j + size, i + size), outline=Color.orange)
    elif checked and not correct:
        # Circle checkbox with red pen.
        try:
            draw.ellipse((j - margin, i - margin, j + size + margin, i + size + margin), width=2, outline=red)
        except TypeError:
            # old PIL versions (<5.1.3)
            draw.ellipse((j - margin, i - margin, j + size + margin, i + size + margin), outline=red)
    elif not checked and correct:
        # Check (cross) the box (with red pen).
        draw.line((j, i, j + size - 1, i + size - 1), fill=red, width=2)
        draw.line((j + size - 1, i, j, i + size - 1), fill=red, width=2)


def _write_score(draw: ImageDraw.ImageDraw, pos: Pixel, earn: float, maximum: float, size: int) -> None:
    i, j = pos
    fnt = ImageFont.truetype("FreeSerif.ttf", int(0.7 * size))
    draw.text((j, i), f"{round(earn, 2):g}/{round(maximum, 2):g}", font=fnt, fill=Color.red)



#!/usr/bin/env python3
import csv
import subprocess
import tempfile
from math import inf
from pathlib import Path
from typing import Union, Literal, Optional

from ptyx_mcq.scan.amend import amend_all
from ptyx_mcq.scan.conflict_solver import ConflictSolver
from ptyx_mcq.scan.data_manager import DataStorage
from ptyx_mcq.scan.document_data import DocumentData, PicData, Page
from ptyx_mcq.scan.pdftools import PIC_EXTS
from ptyx_mcq.scan.scan_pic import (
    scan_picture,
    CalibrationError,
)
from ptyx_mcq.scan.scores_manager import ScoresManager
from ptyx_mcq.tools.config_parser import (
    StudentId,
    StudentName,
    DocumentId,
)
from ptyx_mcq.tools.io_tools import print_success, print_warning, ANSI_RESET, ANSI_GREEN


# -----------------------------------------
#                  Scan
#     Extract info from digitized documents
# -----------------------------------------
#    PTYX
#    Python LaTeX preprocessor
#    Copyright (C) 2009-2020  Nicolas Pourcelot
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


# File `compilation.py` is in ../.., so we have to "hack" `sys.path` a bit.
# script_path = dirname(abspath(sys._getframe().f_code.co_filename))
# sys.path.insert(0, join(script_path, '../..'))
# from ptyx.compilation import join_files, compile_latex

# from ..make.header import answers_and_score


class MissingQuestion(RuntimeError):
    """Error raised when some questions where not seen when scanning all data."""


class MissingConfigurationData(RuntimeError):
    """Error raised when some configuration data is missing."""


class MCQPictureParser:
    """Main class for parsing pdf files containing all the scanned MCQ."""

    def __init__(
        self,
        path: Union[str, Path],
        input_dir: Optional[Path] = None,
        output_dir: Optional[Path] = None,
    ):
        # Set `already_seen` will contain all seen (ID, page) couples.
        # It is used to catch a hypothetical scanning problem:
        # we have to be sure that the same page on the same test is not seen
        # twice.
        self.already_seen: set[tuple[DocumentId, Page]] = set()
        self.warnings = False
        self.data_storage = DataStorage(Path(path), input_dir=input_dir, output_dir=output_dir)
        self.scores_manager = ScoresManager(self)

    @property
    def config(self):
        return self.data_storage.config

    @property
    def data(self):
        return self.data_storage.data

    def _test_integrity(self) -> None:
        """For every test:
        - all pages must have been scanned,
        - all questions must have been seen."""
        questions_not_seen = {}
        pages_not_seen = {}
        ordering = self.config.ordering
        for doc_id in self.data:
            try:
                doc_ordering = ordering[doc_id]
            except KeyError:
                raise MissingConfigurationData(
                    f"No configuration data found for document #{doc_id}.\n"
                    "Maybe you recompiled the ptyx file in the while ?\n"
                    f"(Launching `mcq make -n {max(self.data)}` might fix it.)"
                )
            questions = set(doc_ordering["questions"])
            diff = questions - set(self.data[doc_id]["answered"])
            if diff:
                questions_not_seen[doc_id] = ", ".join(str(q) for q in diff)
            # All tests may not have the same number of pages, since
            # page breaking will occur at a different place for each test.
            pages = set(self.config.boxes[doc_id])
            diff = pages - set(self.data[doc_id]["pages"])
            if diff:
                pages_not_seen[doc_id] = ", ".join(str(p) for p in diff)
        if pages_not_seen:
            self._warn("= WARNING =")
            self._warn("Pages not seen:")
            for doc_id in sorted(pages_not_seen):
                self._warn(f"    • Test {doc_id}: page(s) {pages_not_seen[doc_id]}")
        if questions_not_seen:
            self._warn("=== ERROR ===")
            self._warn("Questions not seen !")
            for doc_id in sorted(questions_not_seen):
                self._warn(f"    • Test {doc_id}: question(s) {questions_not_seen[doc_id]}")

        if questions_not_seen:
            # Don't raise an error for pages not found (only a warning in log)
            # if all questions were found, this was probably empty pages.
            raise MissingQuestion("Questions not seen ! (Look at message above).")

    def _keep_previous_version(self, pic_data: PicData) -> bool:
        """Test if a previous version of the same page exist.

        If so, it probably means the page has been scanned twice, but it could
        also indicate a more serious problem (for example, tests with the same ID
        have been given to different students !).
        As a precaution, we should signal the problem to the user, and ask him
        what he wants to do.
        """
        doc_id = pic_data.doc_id
        p = pic_data.page

        # This page has never been seen before, everything is OK.
        if (doc_id, p) not in self.already_seen:
            self.already_seen.add((doc_id, p))
            return False

        # This is problematic: it seems like the same page has been seen twice.
        lastpic_path = pic_data.pic_path
        lastpic = self.data_storage.relative_pic_path(lastpic_path)
        firstpic_path = self.data[doc_id]["pages"][p].pic_path
        firstpic = self.data_storage.relative_pic_path(firstpic_path)
        assert isinstance(lastpic_path, str)
        assert isinstance(firstpic_path, str)

        self._warn(f"WARNING: Page {p} of test #{doc_id} seen twice " f'(in "{firstpic}" and "{lastpic}") !')
        action = None
        keys = ("name", "student_ID", "answered")
        if all(pic_data[key] == self.data[doc_id]["pages"][p][key] for key in keys):  # type: ignore
            # Same information found on the two pages, just keep one version.
            action = "f"
            self._warn("Both page have the same information, keeping only first one...")

        # We have a problem: this is a duplicate.
        # In other words, we have 2 different versions of the same page.
        # Ask the user what to do.
        while action not in ("f", "l"):
            print("What must we do ?")
            print("- See pictures (s)")
            print("- Keep only first one (f)")
            print("- Keep only last one (l)")

            action = input("Answer:")
            if action == "s":
                with tempfile.TemporaryDirectory() as tmpdirname:
                    path = Path(tmpdirname) / "test.png"
                    # https://stackoverflow.com/questions/39141694/how-to-display-multiple-images-in-unix-command-line
                    subprocess.run(["convert", firstpic_path, lastpic_path, "-append", str(path)], check=True)
                    subprocess.run(["feh", "-F", str(path)], check=True)
                    input("-- pause --")
        # We must memorize which version should be skipped in case user
        # launch scan another time.
        skipped_pic = firstpic if action == "l" else lastpic
        self.data_storage.store_skipped_pic(skipped_pic)

        if action == "l":
            # Remove first picture information.
            del self.data[doc_id]["pages"][p]
            self.data_storage.store_doc_data(firstpic.parent.name, doc_id, p)

        return action == "f"

    def generate_report(self) -> None:
        """Generate CSV files with some information concerning each student.

        Used mainly for debugging.
        """
        # Generate CSV file with ID and pictures names for all students.
        info_path = self.data_storage.files.infos
        info = [
            (
                doc_data["name"],
                doc_data["student_ID"],
                doc_id,
                doc_data["score"],
                [doc_data["pages"][p].pic_path for p in doc_data["pages"]],
            )
            for doc_id, doc_data in self.data.items()
        ]

        with open(info_path, "w", newline="") as csvfile:
            writerow = csv.writer(csvfile).writerow
            writerow(("Name", "Student ID", "Test ID", "Score", "Pictures"))
            for name, student_ID, doc_id, score, paths in sorted(info):
                paths_as_str = ", ".join(str(pth) for pth in paths)
                writerow([name, student_ID, f"#{doc_id}", score, paths_as_str])
        print(f'Infos stored in "{info_path}"\n')

    def generate_amended_pdf(self) -> None:
        amend_all(self.data_storage)

    def scan_picture(self, picture: Union[str, Path], manual_verification: bool = True) -> None:
        """This is used for debuging (it allows to test pages one by one)."""
        # f1-pic-003.jpg (page 25)
        # f12-pic-005.jpg
        # f12-pic-003.jpg
        # f12-pic-004.jpg
        # f12-pic-013.jpg
        # f7-pic-013.jpg
        # f9-pic-004.jpg
        # f9-pic-005.jpg
        # f13-pic-002.jpg
        if not any(str(picture).endswith(ext) for ext in PIC_EXTS):
            raise TypeError("Allowed picture extensions: " + ", ".join(PIC_EXTS))
        pic_path = Path(picture).expanduser().resolve()
        if not pic_path.is_file():
            pic_path = self.data_storage.absolute_pic_path(picture)
        pic_data, _ = scan_picture(pic_path, self.config, manual_verification=manual_verification)
        print(pic_data)

    def _warn(self, *values, sep=" ", end="\n") -> None:
        """Print to stdout and write to log file."""
        msg = sep.join(str(val) for val in values) + end
        print_warning(msg)
        self.data_storage.write_log(msg)
        self.warnings = True

    def scan_all(
        self,
        start: int = 1,
        end: Union[int, float] = inf,
        manual_verification: Optional[bool] = None,
        ask_for_name: bool = False,
        reset: bool = False,
    ) -> None:
        """Extract information from pdf, calculate scores and annotate documents
        to display correct answers."""

        # Test if the PDF files of the input directory have changed and
        # extract the images from the PDF files if needed.
        print("Search for previous data...")
        self.data_storage.reload(reset=reset)

        # Dict `data` will collect data from all scanned tests.
        # ...............................................................
        # FORMAT: {ID: {'pages': (dict) the pages seen, and all related information,
        #               'answers': ({int: set}) the answers of the student for each question,
        #               'score': (float) the test score,
        #               'name': (str) the student name,
        #               'last_pic': (str) last image seen full path,
        #               },
        #           ...
        #          }
        # ...............................................................
        #

        # ---------------------------------------
        # Extract informations from the pictures.
        # ---------------------------------------
        data = self.data
        # self.name2docID = {d["name"]: ID for ID, d in data.items()}
        self.already_seen = set((ID, p) for ID, d in data.items() for p in d["pages"])
        pic_list = self.data_storage.get_pics_list()

        # assert all(isinstance(path, Path) for path in self.data_storage.skipped)
        # assert all(isinstance(path, Path) for path in self.data_storage.verified)

        for i, pic_path in enumerate(pic_list, start=1):
            if not (start <= i <= end):
                continue
            # Make pic_path relative, so that folder may be moved if needed.
            pic_path = self.data_storage.relative_pic_path(pic_path)
            if pic_path in self.data_storage.skipped:
                continue
            print("-------------------------------------------------------")
            print("Page", i)
            print("File:", pic_path)

            # 1) Extract all the data of an image
            #    ‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾

            try:
                # Warning: manual_verification can be None, so the order is important
                # below : False and None -> False (but None and False -> None).
                manual_verification = (pic_path not in self.data_storage.verified) and manual_verification
                pic_data, matrix = scan_picture(
                    self.data_storage.absolute_pic_path(pic_path), self.config, manual_verification
                )
                # `pic_data` FORMAT is specified in `scan_pic.py`.
                # (Search for `pic_data =` in `scan_pic.py`).
                pic_data.pic_path = str(pic_path)
                print()

            except CalibrationError:
                self._warn(f"WARNING: {pic_path} seems invalid ! Skipping...")
                input("-- PAUSE --")
                self.data_storage.store_skipped_pic(pic_path)
                continue

            # if pic_data.verified:
            #     # If the page has been manually verified, keep track of it,
            #     # so it won't be verified next time if a second pass is needed.
            #     self.data_storage.store_verified_pic(pic_path)

            doc_id = pic_data.doc_id
            page = pic_data.page

            if self._keep_previous_version(pic_data):
                continue

            # 2) Gather data
            #    ‾‾‾‾‾‾‾‾‾‾‾
            name, student_ID = self.data_storage.more_infos.get(doc_id, (StudentName(""), StudentId("")))
            doc_data: DocumentData = data.setdefault(
                doc_id,
                DocumentData(
                    pages={}, name=name, student_ID=student_ID, answered={}, score=0, score_per_question={}
                ),
            )
            doc_data["pages"][page] = pic_data

            for q in pic_data.answered:
                ans = doc_data["answered"].setdefault(q, set())
                ans |= pic_data.answered[q]
                # Simplify: doc_data["answered"][q] = set(pic_data["answered"][q])

            # 3) 1st page of the test => retrieve the student name
            #    ‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾
            # if page == 1:
            #     self._extract_name(doc_id, doc_data, matrix)

            if page == 1:
                if doc_id in self.data_storage.more_infos:
                    doc_data["name"], doc_data["student_ID"] = self.data_storage.more_infos[doc_id]
                else:
                    doc_data["name"] = pic_data.name

            # Store work in progress, so we can resume process if something fails...
            self.data_storage.store_doc_data(pic_path.parent.name, doc_id, page, matrix)

        print("Scan successful.")

        # ---------------------------
        # Resolve detected problems
        # ---------------------------
        # Resolve conflicts manually: unknown student ID, ambiguous answer...
        print("\nAnalyzing collected data:")
        ConflictSolver(self.data_storage).resolve_conflicts()

        # ---------------------------
        # Test integrity
        # ---------------------------
        # For every test:
        # - all pages must have been scanned,
        # - all questions must have been seen.
        print("Test for data integrity...")
        self._test_integrity()
        print("Everything seems OK.")

        # ---------------------------
        # Calculate scores
        # ---------------------------
        # Calculate the score, taking care of the chosen mode.
        self.scores_manager.calculate_scores()
        self.scores_manager.print_scores()

        # ---------------------------------------------------
        # Time to synthesize & store all those informations !
        # ---------------------------------------------------
        self.scores_manager.generate_csv_file()
        self.scores_manager.generate_xlsx_file()
        cfg_ext = ".ptyx.mcq.config.json"
        cfg_path = str(self.data_storage.paths.configfile)
        assert cfg_path.endswith(cfg_ext)
        xlsx_symlink = Path(cfg_path[: -len(cfg_ext)] + ".scores.xlsx")
        xlsx_symlink.unlink(missing_ok=True)
        xlsx_symlink.symlink_to(self.data_storage.files.xlsx_scores)
        self.generate_report()
        self.generate_amended_pdf()
        print(f"\n{ANSI_GREEN}Success ! {ANSI_RESET}:)")


def scan(
    path: Path,
    reset: bool = False,
    ask_for_name: bool = False,
    verify: Literal["auto", "always", "never"] = "auto",
) -> None:
    """Implement `mcq scan` command."""
    if verify == "always":
        manual_verification: Optional[bool] = True
    elif verify == "never":
        manual_verification = False
    else:
        manual_verification = None
    MCQPictureParser(path).scan_all(
        reset=reset,
        ask_for_name=ask_for_name,
        manual_verification=manual_verification,
    )
    print_success("Students' marks successfully generated. :)")



import re
from enum import auto, Enum
from functools import partial
from pathlib import Path
from typing import Match

from ptyx_mcq.tools.io_tools import print_info, print_warning, print_error


class IncludeStatus(Enum):
    OK = auto()
    DISABLED = auto()
    NOT_FOUND = auto()
    AUTOMATICALLY_ADDED = auto()

    def __repr__(self):
        return f"{self.__class__.__name__}.{self.name}"


class IncludeParser:
    """Parser used to include files in a ptyx file.

    Ptyx-mcq accept the following syntax to include a file:

        -- path/to/file

    There must be at least one space after the two dashed.

    To disable an include, add a `!` just before the dashes:

        !-- path/to/file

    (You may also simply comment the line using the `#` character with a space after,
    which is the usual syntax to comment pTyX code, but this is less convenient).

    By default, when relative, paths to files refer to the directory where the ptyx file
    is located.

    The following syntax allows to change the directory where the files are searched.
    -- ROOT: /path/to/main/directory
    This will change the search directory for every subsequent path, at least
    until another `-- ROOT:` directive occurs (search directory may be changed
    several times).
    """

    # Store all include paths.
    # Format: {ROOT folder: {relative path: include enabled (True)|disabled (False)|invalid (None)}}
    includes: dict[str, dict[str, IncludeStatus]]
    _root: Path

    def __init__(self, default_path: Path):
        """Path `default_path` should be the parent directory of the ptyx file."""
        self.default_path = default_path
        self._reset()

    def _reset(self):
        self._root = self.default_path
        self._last_status = IncludeStatus.OK
        self.includes = {}

    def _parse_include(self, match: Match, strict: bool) -> str:
        include_enabled = match.group(0)[0] != "!"
        if match.group(0)[0] == "#":
            # Special `#--` comments are used to indicate include status.
            # (Currently, this only used to mark that an include was automatically added.)
            try:
                self._last_status = IncludeStatus[match.group(1).strip().rstrip(":")]
            except KeyError:
                print_warning(f"Invalid line: {match.group(0)}")
            return ""
        status = self._last_status
        self._last_status = IncludeStatus.OK
        pattern = match.group(1).strip()
        if pattern.startswith("ROOT:"):
            if include_enabled:
                path = Path(pattern[5:].strip()).expanduser()
                if not path.is_absolute():
                    path = (self.default_path / path).resolve()
                print_info(f"Directory for files inclusion set to '{path}'.")
                if not path.is_dir():
                    raise FileNotFoundError(
                        f"Directory '{path}' not found.\n"
                        f'HINT: Change "-- {pattern}" line in your ptyx file.'
                    )
                self._root = path
            return ""
        else:
            includes = self.includes.setdefault(str(self._root), {})
            if include_enabled:
                file_found = False
                contents = []
                for path in sorted(self._root.glob(pattern)):
                    if path.is_file():
                        file_found = True
                        contents.append(self._include_file(path).strip())
                if file_found:
                    # Include enabled
                    includes[str(pattern)] = status
                else:
                    # Invalid include
                    includes[str(pattern)] = IncludeStatus.NOT_FOUND
                    if strict:
                        print_error(message := f"No file corresponding to {pattern!r} in '{self._root}'!")
                        raise FileNotFoundError(message)
                    else:
                        print_warning(f"No file corresponding to {pattern!r} in '{self._root}'!")
                return "\n\n" + "\n\n".join(contents) + "\n\n"
            else:
                # Include disabled
                includes[str(pattern)] = IncludeStatus.DISABLED
                return ""

    def _include_file(self, path: Path) -> str:
        lines: list[str] = []
        with open(path) as file:
            file_content = file.read().strip()
            # Remove comments
            file_content = re.sub("( # .+)|(^# .+\n)", "", file_content, flags=re.MULTILINE)
            # Each exercise must start with a star. Since each included file is supposed to be an exercise,
            # add the initial star if missing.
            if file_content[:2].strip() != "*":
                file_content = "*\n" + file_content
            for line in file_content.split("\n"):
                lines.append(line)
                if (
                    line.startswith("* ")
                    or line.startswith("> ")
                    or line.startswith("OR ")
                    or line.rstrip() in ("*", ">", "OR")
                ):
                    prettified_path = path.parent / f"\u001b[36m{path.name.replace('#', '##')}\u001b[0m"
                    lines.append(f'#PRINT{{\u001b[36mIMPORTING\u001b[0m "{prettified_path}"}}')
        return "\n".join(lines)

    def parse(self, code: str, strict: bool = False) -> str:
        """Parse all include directives and include the corresponding files into the code.

        By default, if a path or a pattern don't match any file, a warning is raised,
        but the parsing will go on.
        However, if `strict` is set to `True`, a `FileNotFoundError` is raised in that case.
        """
        self._reset()
        _parse = partial(self._parse_include, strict=strict)
        return re.sub(r"^[#!]?-- (.+)$", _parse, code, flags=re.MULTILINE)

    def update(self, ptyxfile_path: Path):
        """Track all the `.ex` files and update pTyX file to include all the `.ex` files found.

        The `.ex` files are searched:
            - in the same directory as the `.pTyX` file
            - in any directory manually added via a `-- ROOT: /my/path` directive.
        """
        self.parse(ptyxfile_path.read_text(encoding="utf8"))
        new_includes: dict[str, dict[str, IncludeStatus]] = {}
        # Several patterns may refer to the same file: keep track of the files already indexed,
        # using their absolute path.
        already_indexed: set[Path] = set()
        for directory, indexed_files in self.includes.items():
            directory_path = Path(directory)
            new_directory_index = new_includes[directory] = {}
            # `index_files` is a dict. Its format is: {
            #   "a relative path or pattern to one or several indexed file": the status of the indexed files
            #   }
            indexed_ex_files: dict[Path, IncludeStatus] = {
                path: status
                for pattern, status in indexed_files.items()
                for path in directory_path.glob(pattern)
                if path not in already_indexed
            }
            already_indexed.update(indexed_ex_files)

            for ex_file_path, status in indexed_ex_files.items():
                ex_file = str(ex_file_path.relative_to(directory_path))
                new_directory_index[ex_file] = status

        for directory, indexed_files in self.includes.items():
            directory_path = Path(directory)
            # Search for new `.ex` files to index:
            for ex_file_path in directory_path.glob("**/*.ex"):
                if ex_file_path not in already_indexed:
                    already_indexed.add(ex_file_path)
                    ex_file = str(ex_file_path.relative_to(directory_path))
                    new_includes[directory][ex_file] = IncludeStatus.AUTOMATICALLY_ADDED
        self.includes = new_includes
        self._rewrite_file(ptyxfile_path, new_includes)

    def _rewrite_file(self, ptyxfile_path: Path, includes: dict[str, dict[str, IncludeStatus]]):
        new_file_added = False
        lines = []
        with open(ptyxfile_path, encoding="utf8") as f:
            for line in f:
                if line.startswith(">>>"):  # `>>>` marks the end of the MCQ
                    # Insert all include directives just before the end of the MCQ.
                    for folder, paths in includes.items():
                        folder_path = Path(folder)
                        if folder_path.is_relative_to(self.default_path):
                            folder = str(folder_path.relative_to(self.default_path))
                        lines.append(f"-- ROOT: {folder}")
                        for path, status in sorted(paths.items()):
                            match status:
                                case IncludeStatus.OK:
                                    lines.append(f"-- {path}")
                                case IncludeStatus.DISABLED:
                                    lines.append(f"!-- {path}")
                                case IncludeStatus.NOT_FOUND:
                                    lines.append(f"#-- NOT_FOUND:\n!-- {path}")
                                    print_warning(f"Invalid file path removed: '{path}'.")
                                case IncludeStatus.AUTOMATICALLY_ADDED:
                                    lines.append(f"#-- AUTOMATICALLY_ADDED:\n-- {path}")
                                    print_info(f"New file added: '{path}'.")
                                    new_file_added = True
                if not re.match(r"[#!]?-- ", line):
                    lines.append(line.rstrip())
        if not new_file_added:
            print_info("No new file found.")
        with open(ptyxfile_path, "w", encoding="utf8") as f:
            f.write("\n".join(lines) + "\n")
